@@ -15,6 +15,13 @@ use std::ffi::c_void;
 use std::mem::size_of;
 use std::sync::atomic::{AtomicPtr, Ordering};
 
+#[cfg(not(feature = "use-bindgen"))]
+pub mod bindings;
+
+#[cfg(not(feature = "use-bindgen"))]
+pub use bindings::*;
+
+#[cfg(feature = "use-bindgen")]
 include!(concat!(env!("OUT_DIR"), "/bindings.rs"));
 
 // define
@@ -365,14 +372,6 @@ pub unsafe fn jl_datatype_nfields(t: *mut jl_datatype_t) -> u32 {
     (&*(&*(t)).layout).nfields
 }
 
-/*
-#define jl_datatype_isinlinealloc(t) (((jl_datatype_t *)(t))->isinlinealloc)
-*/
-#[inline(always)]
-pub unsafe fn jl_datatype_isinlinealloc(t: *mut jl_datatype_t) -> u8 {
-    (&*(t)).isinlinealloc
-}
-
 /*
 #define jl_symbol_name(s) jl_symbol_name_(s)
 */
@@ -440,14 +439,6 @@ pub unsafe fn jl_is_datatype(v: *mut jl_value_t) -> bool {
 */
 // Not implemented
 
-/*
-#define jl_is_immutable(t)   (!((jl_datatype_t*)t)->mutabl)
-*/
-#[inline(always)]
-pub unsafe fn jl_is_immutable(v: *mut jl_value_t) -> bool {
-    (&*jl_typeof(v).cast::<jl_datatype_t>()).mutabl == 0
-}
-
 /*
 #define jl_is_immutable_datatype(t) (jl_is_datatype(t) && (!((jl_datatype_t*)t)->mutabl))
 */
@@ -696,19 +687,6 @@ pub unsafe fn jl_array_isbitsunion(a: *mut jl_array_t) -> bool {
 */
 // Not implemented
 
-/*
-#define julia_init julia_init__threading
-*/
-// Not implemented
-
-/*
-#define jl_init jl_init__threading
-*/
-#[inline(always)]
-pub unsafe fn jl_init() {
-    jl_init__threading()
-}
-
 /*
 #define jl_init_with_image jl_init_with_image__threading
 */
@@ -858,6 +836,16 @@ STATIC_INLINE void jl_array_uint8_set(void *a, size_t i, uint8_t x) JL_NOTSAFEPO
 */
 // Not implemented
 
+#[inline(always)]
+pub unsafe fn jl_is_immutable(v: *mut jl_value_t) -> bool {
+    jl_typeof(v)
+        .cast::<jl_datatype_t>()
+        .as_ref()
+        .unwrap()
+        .mutabl
+        == 0
+}
+
 /*STATIC_INLINE jl_svec_t *jl_field_names(jl_datatype_t *st) JL_NOTSAFEPOINT
 {
     jl_svec_t *names = st->names;
@@ -867,14 +855,8 @@ STATIC_INLINE void jl_array_uint8_set(void *a, size_t i, uint8_t x) JL_NOTSAFEPO
 }*/
 #[inline]
 pub unsafe fn jl_field_names(st: *mut jl_datatype_t) -> *mut jl_svec_t {
-    let mut st = &mut *st;
-    if !st.names.is_null() {
-        return st.names;
-    }
-
-    st.names = (&*st.name).names;
-
-    return st.names;
+    let st = &mut *st;
+    return (&*st.name).names;
 }
 
 /*
@@ -970,7 +952,12 @@ STATIC_INLINE int jl_is_structtype(void *v) JL_NOTSAFEPOINT
 pub unsafe fn jl_is_structtype(v: *mut jl_value_t) -> bool {
     jl_is_datatype(v)
         && jl_is_immutable(v)
-        && (&*v.cast::<jl_datatype_t>()).abstract_ == 0
+        && jl_typeof(v)
+            .cast::<jl_datatype_t>()
+            .as_ref()
+            .unwrap()
+            .abstract_
+            == 0
         && jl_datatype_nfields(v.cast()) == 0
         && jl_datatype_size(v.cast()) > 0
 }
@@ -1005,7 +992,13 @@ STATIC_INLINE int jl_is_abstracttype(void *v) JL_NOTSAFEPOINT
 */
 #[inline]
 pub unsafe fn jl_is_abstracttype(v: *mut c_void) -> bool {
-    jl_is_datatype(v.cast()) && (&*v.cast::<jl_datatype_t>()).abstract_ > 0
+    jl_is_datatype(v.cast())
+        && jl_typeof(v.cast())
+            .cast::<jl_datatype_t>()
+            .as_ref()
+            .unwrap()
+            .mutabl
+            != 0
 }
 
 /*
@@ -1351,6 +1344,11 @@ pub unsafe fn jl_array_dims<'a>(array: *mut jl_array_t, ndims: usize) -> &'a [us
     std::slice::from_raw_parts(x, ndims)
 }
 
+#[inline(always)]
+pub unsafe fn jl_array_dims_ptr<'a>(array: *mut jl_array_t) -> *mut usize {
+    &mut (&mut *array).nrows
+}
+
 /*#define jl_array_ptr_data(a)  ((jl_value_t**)((jl_array_t*)(a))->data)
 STATIC_INLINE jl_value_t *jl_array_ptr_ref(void *a JL_PROPAGATES_ROOT, size_t i) JL_NOTSAFEPOINT
 {
@@ -1392,6 +1390,17 @@ pub unsafe fn jl_array_ptr_set(a: *mut c_void, i: usize, x: *mut c_void) -> *mut
     x.cast()
 }
 
+pub unsafe fn jl_init() {
+    jl_init__threading()
+}
+
+pub unsafe fn jl_init_with_image(
+    julia_bindir: *const ::std::os::raw::c_char,
+    image_relative_path: *const ::std::os::raw::c_char,
+) {
+    jl_init_with_image__threading(julia_bindir, image_relative_path)
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
